'''
This module is a (mostly) redis-py compatible library.
* Unlike redis-py, the backend is configurable, it could write to a Redis, it
  could save the data locally.
* It generates resp command objects that are run by the backend
* It doesn't return values for incrby so incr, incrby, decr, and decrby are pure writes.
* Append doesn't return a string length; same reasoning as incrby.

Reference API: https://github.com/andymccurdy/redis-py/blob/master/redis/client.py
'''

from seneca.engine.storage.redisnap.commands import *
from seneca.engine.util import grouper

import seneca.engine.storage.redisnap.resp_types as resp_types

class Client:
    '''
    Implementation of the API provided by redis-py's StrictRedis
    '''

    def __init__(self, executer):
        self.execute_command = executer

    def _purge(self):
        return self.execute_command.purge()

    def exists(self, name):
        """
        Returns a boolean indicating whether key ``name`` exists
        >>> s._purge()
        >>> s.exists('foo')
        False
        """
        return self.execute_command(Exists(name))
    __contains__ = exists

    def type(self, name):
        """
        Returns the type of key ``name``
        >>> s._purge()
        >>> s.type('foo')
        b'none'
        """
        rtype = self.execute_command(Type(name))

        if issubclass(rtype, RDoesNotExist):
            return str.encode('none')
        elif issubclass(rtype, RScalar):
            return str.encode('string')
        elif issubclass(rtype, RHash):
            return str.encode('hash')
        else:
            raise NotImplementedError()

    def append_nr(self, key, value):
        """
        Appends the string ``value`` to the value at ``key``. If ``key``
        doesn't already exist, create it with a value of ``value``.
        Returns the new length of the value at ``key``.

        >>> s.append('foo', 'bar')
        >>> s.get('foo')
        b'bar'

        """
        return self.execute_command(Append(key, value))


    def get(self, name):
        """
        >>> s._purge()

        Return the value at key ``name``, or None if the key doesn't exist
        >>> s.get('foo') is None
        True
        """
        # TODO: Decide how we want to handle non-existing keys in the commands api
        ret = self.execute_command(Get(name))
        if isinstance(ret, RDoesNotExist):
            return None
        elif isinstance(ret, RScalar):
            return ret.to_bytes()
        else:
            raise Exception('Something went wrong.')


    def __getitem__(self, name):
        """
        Return the value at key ``name``, raises a KeyError if the key
        doesn't exist.
        """
        value = self.get(name)
        if value is not None:
            return value
        raise KeyError(name)


    def set_nr(self, name, value, ex=None, px=None, nx=False, xx=False):
        """
        >>> s._purge()

        >>> s.set_nr('foo', 'bar'); s.get('foo')
        True
        b'bar'

        >>> s.set_nr('foo', 'baz'); s.get('foo')
        True
        b'baz'

        Set the value at key ``name`` to ``value``
        ``ex`` sets an expire flag on key ``name`` for ``ex`` seconds.
        ``px`` sets an expire flag on key ``name`` for ``px`` milliseconds.
        ``nx`` if set to True, set the value at key ``name`` to ``value`` only
            if it does not exist.
        ``xx`` if set to True, set the value at key ``name`` to ``value`` only
            if it already exists.
        """
        assert ex is None, 'Cache expiration not supported'
        assert px is None, 'Cache expiration not supported'
        assert nx is False # TODO: Will add this later
        assert nx is False # TODO: Will add this later

        self.execute_command(SetNR(name, value))

        return True


    def __setitem__(self, name, value):
        """
        >>> s._purge()
        >>> s['foo'] = 'bar'
        """
        self.set_nr(name, value)


    def incr_nr(self, name, amount=1):
        """
        >>> s._purge() is None
        True

        Increments the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as ``amount``
        >>> s.incr_nr('foo', 1) is None
        True
        """
        self.execute_command(IncrByNR(name, amount))


    def incrby_nr(self, name, amount=1):
        """
        >>> s._purge()

        Increments the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as ``amount``
        >>> s.incrby_nr('foo', 1) is None
        True
        """
        self.execute_command(IncrByNR(name, amount))

    def append_nr(self, name, value):
        """
        Increments the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as ``amount``
        >>> s._purge()
        >>> s.append_nr('foo', 1) is None
        True
        """
        self.execute_command(AppendNR(name, value))


    def decr_nr(self, name, amount=1):
        """
        Decrements the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as 0 - ``amount``
        >>> s.decr_nr('foo', 1) is None
        True
        """
        self.execute_command(IncrByNR(name, 0 - amount))

    def hget(self, name, key):
        """
        >>> s._purge()
        >>> s.hget('foo', 'bar') is None
        True
        """
        ret = self.execute_command(HGet(name, key))
        if isinstance(ret, RDoesNotExist):
            return None
        elif isinstance(ret, RScalar):
            return ret.to_bytes()

    def hset_nr(self, name, key, value):
        """
        >>> s.hset_nr('foo', 'bar', 'baz'); s.hget('foo', 'bar')
        b'baz'
        """
        self.execute_command(HSetNR(name, key, value))


    def hexists(self, name, key):
        """
        >>> s._purge()
        >>> s.hexists('foo', 'bar')
        False

        >>> s.hset_nr('foo', 'bar', 'baz')

        >>> s.hexists('foo', 'qux')
        False

        >>> s.hexists('foo', 'bar')
        True
        """
        return self.execute_command(HExists(name, key))


    def lindex(self, name, index):
        """
        Tested in lpush_nr
        """
        return self.execute_command(LIndex(name, index)).to_bytes()



    def lpush_nr(self, name, *values):
        """
        >>> s._purge()
        >>> s.lpush_nr('foo', 'bar')

        >>> s.lindex('foo', 0)
        b'bar'

        >>> s.lindex('foo', 9)

        """
        self.execute_command(LPushNR(name, values))



    def lset(self, name, index, value):
        """
        >>> s._purge()
        >>> s.lpush_nr('foo', 'bar')
        >>> s.lset('foo', 0, 'bar')
        True

        """
        self.execute_command(LSet(name, index, value))
        return True


    def rpush_nr(self, name, *values):
        """
        >>> s._purge()
        >>> s.rpush_nr('foo', 'bar')

        >>> s.lindex('foo', 0)
        b'bar'

        """
        self.execute_command(RPushNR(name, values))


    def lpop(self, name):
        """
        >>> s._purge()
        >>> s.rpush_nr('foo', 'bar')

        >>> s.lpop('foo')
        b'bar'
        """
        return self.execute_command(LPop(name)).to_bytes()


    def rpop(self, name):
        """
        >>> s._purge()
        >>> s.rpush_nr('foo', 'bar')

        >>> s.rpop('foo')
        b'bar'
        """
        return self.execute_command(RPop(name)).to_bytes()

    def zadd_nr(self, name, *member_score_pairs, **member_score_as_kwargs):
        """
        Tested in zscore
        """
        member_scores_all = dict(grouper(member_score_pairs, 2))
        member_scores_all.update(member_score_as_kwargs)
        self.execute_command(ZAddNR(name, member_scores_all))

    def zscore(self, name, value):
        """
        >>> s._purge()
        >>> s.zadd_nr('foo', 'foo', 1, 'bar', 2, baz=3)

        >>> s.zscore('foo', 'foo')
        1
        >>> s.zscore('foo', 'bar')
        2
        >>> s.zscore('foo', 'baz')
        3
        """
        return self.execute_command(ZScore(name, value))


    def zrem_nr(self, name, *values):
        """
        >>> s._purge()
        >>> s.zadd_nr('foo', foo=1, bar=2, baz=3)
        >>> s.zrem_nr('foo', 'foo', 'bar', 'baz')
        >>> s.zscore('foo', 'foo'); s.zscore('foo', 'bar'); s.zscore('foo', 'baz');
        """
        self.execute_command(ZRemNR(name, values))


    def zrevrangebyscore(self, name, max, min, withscores=False):
        """
        >>> s._purge()
        >>> s.zadd_nr('foo', foo=1, bar=2, baz=3, qux=4, quux=5)
        >>> s.zrevrangebyscore('foo', 4, 2)
        ['qux', 'baz', 'bar']

        >>> s.zrevrangebyscore('foo', 4, 2, withscores=True)
        [(4, 'qux'), (3, 'baz'), (2, 'bar')]
        """
        # TODO: Consider not constructing list, will break API a bit, but likely more performant in some cases.
        return list(self.execute_command(ZRevRangeByScore(name, max, min, with_scores=withscores)))

    def zincrby_nr(self, name, value, amount=1):
        """
        >>> s._purge()
        >>> s.zincrby_nr('foo', 'bar', 2)
        >>> s.zscore('foo', 'bar')
        2

        >>> s.zincrby_nr('foo', 'bar')
        >>> s.zscore('foo', 'bar')
        3
        """
        self.execute_command(ZIncrByNR(name, amount, value))


def run_tests(deps_provider):
    '''
    '''
    import seneca.engine.storage.redisnap.local_backend as l_back

    c = Client(executer = print)
    s = Client(executer = l_back.Executer())

    import doctest, sys
    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



from seneca.engine.util import *
from seneca.engine.storage.mysql_intermediate import *

dissallowed_queries = [ AddTableColumn,
                        DropTableColumn,
                        DeleteRows,
                        UpdateRows,
                        InsertRows,
                        DropTable,
                        CreateTable,
                        SetRows,
]

class Executer(object):
    @auto_set_fields
    def __init__(self, ex):
        pass

    def __call__(self, query):
        type(query)
        assert type(query) not in dissallowed_queries, 'Queries that modify data are not allowed in mysql_ro_executer read only context.'
        return self.ex(query)

    def many(self, queries):
        for q in queries:
            self(q)

    def kill(self):
        return self.ex.kill()


def run_tests(deps_provider):
    pass
    '''
    Clear DB:
    >>> _ = bex.cur.execute('DROP DATABASE seneca_test;')
    >>> _ = bex.cur.execute('CREATE DATABASE seneca_test;')
    >>> _ = bex.cur.execute('use seneca_test;')

    >>> ro_ex = ExecuterBase(bex)

    >>> try_ex_catch(ro_ex, DeleteRows('test_users', QueryCriterion('eq', 'username', 'test')))
    Queries that modify data are not allowed in mysql_ro_executer read only context.
    >>> try_ex_catch(ro_ex, UpdateRows( 'test_users',
    ...                  QueryCriterion('eq', 'username', 'tester'),
    ...                  {'balance': 0, 'status':'broke'}))
    Queries that modify data are not allowed in mysql_ro_executer read only context.
    >>> try_ex_catch(ro_ex, SetRows( 'test_users',
    ...                  [[0, 'broke']]))
    Queries that modify data are not allowed in mysql_ro_executer read only context.
    >>> try_ex_catch(ro_ex, InsertRows('test_users', ['username', 'first_name', 'balance'],
    ...   [['tester', 'test', 500],
    ...    ['tester2', 'two', 200],
    ...   ]))
    Queries that modify data are not allowed in mysql_ro_executer read only context.
    >>> try_ex_catch(ro_ex, InsertRows('test_users', ['username', 'first_name', 'balance'],[['tester', 'test', 500]]))
    Queries that modify data are not allowed in mysql_ro_executer read only context.
    >>> try_ex_catch(ro_ex, DropTableColumn('test_users', 'balance2'))
    Queries that modify data are not allowed in mysql_ro_executer read only context.
    >>> try_ex_catch(ro_ex, DropTable('test_users'))
    Queries that modify data are not allowed in mysql_ro_executer read only context.
    >>> try_ex_catch(ro_ex, ct)
    Queries that modify data are not allowed in mysql_ro_executer read only context.

    Setup some data:
    >>> print(bex(ct))
    SQLExecutionResult({'success': True, 'data': None})

    >>> print(bex(InsertRows('test_users', ['username'],
    ...   [['user_a'],
    ...    ['user_b'],
    ...   ])))
    SQLExecutionResult({'success': True, 'data': {'last_row_id': 1, 'row_count': 2}})

    >>> ro_ex(SelectRows('test_users', [], None, None, None)).success
    True
    >>> ro_ex(CountUniqueRows('test_users', 'id', None)).success
    True
    >>> ro_ex(CountRows('test_users', None)).success
    True
    >>> ro_ex(DescribeTable('test_users')).success
    True
    >>> ro_ex(ListTables()).success
    True
    '''
    import doctest, sys
    from seneca.engine.storage.mysql_executer import Executer
    bex = deps_provider(Executer)

    ct = CreateTable('test_users',
                     AutoIncrementColumn('id'),
                     [ ColumnDefinition('username', SQLType('VARCHAR', 30), True)]
                    )

    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



'''
TODO: complex stuff, soft delete a permanent, create temporary, delete temporary
TODO: forbid queries with names matching temp tables

TODO: Currently circumventing main code block that selects special query actions
'''
import re
import copy
import MySQLdb
from seneca.engine.util import *

import seneca.engine.storage.mysql_executer as ex_base
from seneca.engine.storage.mysql_intermediate import *

unalterted_queries = [ DeleteRows,
                       UpdateRows,
                       InsertRows,
                       SelectRows,
                       CountUniqueRows,
                       CountRows,
                       DescribeTable,
]

dissallowed_queries = [ AddTableColumn,
                        DropTableColumn,
]

def make_temp_name(table_name):
    '''
    >>> make_temp_name('test')
    '$temp$test'
    '''
    return '$temp$' + table_name


def make_query_with_temp_name(q):
    '''
    >>> q = mysqli.InsertRows('test_users', ['username', 'first_name', 'balance'],
    ...   [['tester', 'test', 500],
    ...    ['tester2', 'two', 200],
    ...   ])
    >>> q1 = make_query_with_temp_name(q)
    >>> q1.table_name
    '$temp$test_users'
    '''
    q1 = copy.copy(q)
    q1.table_name = make_temp_name(q1.table_name)
    return q1


def CreateTableAction(ex, q):
    '''
    Clear state:
    >>> _ = ex.rollback()
    >>> _ = ex.cur.execute('DROP DATABASE seneca_test;')
    >>> _ = ex.cur.execute('CREATE DATABASE seneca_test;')
    >>> _ = ex.cur.execute('use seneca_test;')

    >>> _ = CreateTableAction(ex, ct)

    Confirm that table name is stored in temp_tables field, used for rollback and commit
    >>> ex.temp_tables
    {'test_users'}

    Confirm that the table is created (not established in this step if it's actually a temp table)
    >>> ex.cur.execute('SELECT * from seneca_test.$temp$test_users')
    0
    >>> print(ex.cur.fetchall())
    ()

    Confirm that the table is temporary (temporary tables are not returned by mysql's show tables)
    >>> ex.cur.execute('SHOW TABLES;')
    0

    Make sure no table has actually been created and commited.
    >>> ist = easy_db.Table.from_existing('INFORMATION_SCHEMA.TABLES').run(ex)
    >>> len(ist.select('table_name').where(ist.TABLE_NAME == 'test_users').run(ex))
    0
    >>> ex.rollback()

    Repeat process but commit:
    >>> _ = CreateTableAction(ex, ct)
    >>> _ = ex.commit()
    >>> ist = easy_db.Table.from_existing('INFORMATION_SCHEMA.TABLES').run(ex)
    >>> len(ist.select('table_name').where(ist.TABLE_NAME == 'test_users').run(ex))
    1
    >>> ex.temp_tables
    set()

    TODO: More unit tests on this.

    Dimensions of scenarios
    * (P)ermanent table exists [True, False]
    * (S)oft delete flag exists [True, False]
    * (T)empoary table exists [True, False]
    * (I)f_not_exists set on query [True, False]

    * P=True, S=False, T=True, I=any -> Exception (two version of table temp and per logically exist, this should never happen)
    * P=False, S=True, T=any, I=any -> Exception (soft delete but no permanent table to be soft deleted, this should never happen)
    * P=True, S=True, T=False, I=any -> run query not as temp query to create table
    ----

    * P=S, T=True, I=True -> run query as temp verify it works correctly with I on a temp table
    * P=S, T=False, I=any -> run a temporary query, create the temp table
    '''

    name = q.table_name
    temp_table_exists = name in ex.temp_tables
    permanent_table_exists = 0 < ex.cur.execute("SELECT table_name FROM INFORMATION_SCHEMA.TABLES where table_name = '{}'".format(name))
    soft_delete_exists = name in ex.soft_deleted_tables
    #q.if_not_exists

    q_type = type(q)
    assert q_type == CreateTable, 'Argument 1 q is wrong type, should be CreateTable got %' % str(q_type)

    if permanent_table_exists and (not soft_delete_exists) and temp_table_exists:
        Exception("SPITS is in an inconsitent state, temp table and permanent table exists, but no soft delete record is present!!!")

    if (not permanent_table_exists) and soft_delete_exists:
        Exception("SPITS is in an inconsitent state, soft delete record present but permanent table missing!!!")

    if permanent_table_exists and (not soft_delete_exists):
        if q.if_not_exists:
            return ex_base.SQLExecutionResult(True, "Permanent table already exists, but okay, if_not_exists is set to True")
        else:
            Exception("SPITS error, permanent table already exists.")
    else:
        # Important, temp tables are stored in executer temp table without temp table prefix
        ex.temp_tables.add(name)
        # Prefix added here for create query
        sql_str = make_query_with_temp_name(q).to_sql()

        # NOTE: A little janky
        sql_preamble_re = re.compile('^CREATE TABLE')
        assert re.match(sql_preamble_re, sql_str)
        new_query = re.sub(sql_preamble_re, 'CREATE TEMPORARY TABLE', sql_str)
        res = ex.cur.execute(new_query)

        return ex_base.format_result(q_type, ex.cur)


def ListTablesAction(ex, q):
    '''
    '''
#
#    TODO: finish this tests
#    Result comes from db:
#    >> _ = ex.commit()
#    >> str(ListTablesAction(ex,ListTables()))
#    "SQLExecutionResult({'success': True, 'data': ['test_users']})"
#
#    Soft delete table hidden:
#    >> _ = CreateTableAction(ex, ct)
#    >> _ = ex.commit()
#
#    '''
    q_type = type(q)
    ex.cur.execute(q.to_sql())

    sql_ex_res = ex_base.format_result(q_type, ex.cur)
    table_names = sql_ex_res.data

    table_names_temp = list(
                         set( ex.temp_tables | \
                              {t for t in table_names if t not in ex.soft_deleted_tables}
                         )
                       )
    sql_ex_res.data = table_names_temp

    # Not formatting
    return ex_base.SQLExecutionResult(True, sql_ex_res.data)


def DropTableAction(ex, q):
    '''
    Clear state:
    >>> _ = ex.rollback()
    >>> _ = ex.cur.execute('DROP DATABASE seneca_test;')
    >>> _ = ex.cur.execute('CREATE DATABASE seneca_test;')
    >>> _ = ex.cur.execute('use seneca_test;')

    Delete an uncommited table (temporary):
    >>> _ = CreateTableAction(ex, ct)
    >>> ex.temp_tables
    {'test_users'}

    >>> str(DropTableAction(ex, DropTable('test_users')))
    "SQLExecutionResult({'success': True, 'data': 'Uncommitted table test_users has been deleted.'})"

    >>> ex.temp_tables
    set()


    Clear state:
    >>> _ = ex.rollback()
    >>> _ = ex.cur.execute('DROP DATABASE seneca_test;')
    >>> _ = ex.cur.execute('CREATE DATABASE seneca_test;')
    >>> _ = ex.cur.execute('use seneca_test;')

    Soft delete a table and confirm name has been added to the soft_delete list:
    >>> _ = CreateTableAction(ex, ct)
    >>> _ = ex.commit()
    >>> str(DropTableAction(ex, DropTable('test_users')))
    "SQLExecutionResult({'success': True, 'data': 'Table test_users has been soft-deleted.'})"
    >>> ex.soft_deleted_tables
    {'test_users'}
    >>> ex.temp_tables
    set()

    >>> _ = ex.commit()

    Test that soft-deleting an already soft deleted table sends a failure:
    >>> str(DropTableAction(ex, DropTable('test_users')))
    'SQLExecutionResult({\\'success\\': False, \\'data\\': "SPITS ERROR: Attempting to delete a table that doesn\\'t logically exist"})'
    '''
    q_type = type(q)

    name = q.table_name
    temp_name = make_temp_name(name)

    temp_table_exists = name in ex.temp_tables
    permanent_table_exists = 0 < ex.cur.execute("SELECT table_name FROM INFORMATION_SCHEMA.TABLES where table_name ='{}'".format(name))
    soft_delete_exists = name in ex.soft_deleted_tables

    if temp_table_exists:
        # TODO: Make sure this is the right logic and in the right place
        assert permanent_table_exists == soft_delete_exists, "SPITS ERROR: Table delete status in inconsistent state."

        ex.cur.execute("DROP TEMPORARY TABLE %s" % temp_name)
        # TODO: validate result
        ex.temp_tables.remove(name)
        return ex_base.SQLExecutionResult(True, "Uncommitted table %s has been deleted." % name)

    elif permanent_table_exists and (not soft_delete_exists):
        ex.soft_deleted_tables.add(name)
        return ex_base.SQLExecutionResult(True, "Table %s has been soft-deleted." % name)

    else:
        return ex_base.SQLExecutionResult(False, 'SPITS ERROR: Attempting to delete a table that doesn\'t logically exist')



special_action_query_dict = { CreateTable: CreateTableAction,
                              ListTables: ListTablesAction,
                              DropTable: DropTableAction,
}

special_action_queries = list(special_action_query_dict.keys())

# TODO: Dedupe with main executer
class Executer(object):
    def __init__(self, username, password, db, host, port=3306):
        ''' Tested below '''
        self.conn = MySQLdb.connect(host=host, user=username, passwd=password,
                                    db=db, port=port)
        self.conn.autocommit = False
        self.cur = self.conn.cursor()
        self.temp_tables = set()
        self.soft_deleted_tables = set()

    def __call__(self, query):
        q_type = type(query)
        assert issubclass(q_type, Query), 'The passed parameter is not a query.'

        try:
            if q_type in unalterted_queries:

                if (query.table_name in self.soft_deleted_tables
                    and query.table_name not in self.temp_tables):
                    raise Exception("Table does not exist, (soft deleted).")

                if query.table_name in self.temp_tables:
                    query = make_query_with_temp_name(query)

                self.cur.execute(query.to_sql())
                return ex_base.format_result(q_type, self.cur)

            elif q_type in dissallowed_queries:
                raise Exception('Dissallowed query. Incompatible with SPITS.')

            elif q_type in special_action_queries:
                return special_action_query_dict[q_type](self, query)

            else:
                raise Exception("Unrecognized query type. \
SPITS does not have the needed info to execute this query.")

        except Exception as err:
            # Note: This function may return a formated result, or it may reraise the error
            return ex_base.handle_error(q_type, err)

    def _clear(self):
        self.temp_tables = set()
        self.soft_deleted_tables = set()


    def commit(self):
        ''' Tested elsewhere '''
        # DML commit
        self.conn.commit()

        # DDL pseudo-commit
        for t in self.temp_tables:
            name = t
            temp_name = make_temp_name(name)

            q_str = 'CREATE TABLE {name} LIKE {temp_name};\
                     INSERT {name} SELECT * FROM {temp_name};\
                     DROP TEMPORARY TABLE {temp_name};'.format(**locals())

            self.cur.execute(q_str)

        for t in self.soft_deleted_tables:
            self.cur.execute("DROP TABLE %s;" % t)

        # Purge internal scratch
        self._clear()


    def rollback(self):
        ''' Tested elsewhere '''
        # DML rollback
        self.conn.rollback()

        # DDL pseudo-rollback
        for t in self.temp_tables:
            self.cur.execute("DROP TEMPORARY TABLE %s;" % make_temp_name(t))

        # Purge internal scratch
        self._clear()

    def many(self, queries):
        # TODO: Test the speed on this
        # TODO: Error handling
        # TODO: Test atomicity
        for q in queries:
            self.cur.execute(q.to_sql())
        self.conn.commit()

        return ex_base.SQLExecutionResult(True, None)

    def kill(self):
        self.cur.close()
        self.conn.close()



def run_tests(deps_provider):
    '''
    Setup/clear state:
    >>> _ = spex.rollback(); spex.soft_deleted_tables |  spex.temp_tables
    set()
    >>> _ = spex.cur.execute('DROP DATABASE IF EXISTS seneca_test;')
    >>> _ = spex.cur.execute('CREATE DATABASE seneca_test;')
    >>> _ = spex.cur.execute('use seneca_test;')

    List empty:
    >>> str(ListTablesAction(spex,ListTables()))
    "SQLExecutionResult({'success': True, 'data': []})"

    Create temporary table:
    >>> _ = CreateTableAction(spex, ct)
    >>> spex.temp_tables
    {'test_users'}

    List with results from temporary table record in executer:
    >>> str(ListTablesAction(spex,ListTables()))
    "SQLExecutionResult({'success': True, 'data': ['test_users']})"

    List with results from permanent table:
    >>> spex.commit()
    >>> spex.temp_tables
    set()
    >>> str(ListTablesAction(spex,ListTables()))
    "SQLExecutionResult({'success': True, 'data': ['test_users']})"

    Results exclude soft deleted table:
    >>> str(DropTableAction(spex, DropTable('test_users')))
    "SQLExecutionResult({'success': True, 'data': 'Table test_users has been soft-deleted.'})"

    Verify soft delete of commited table:
    >>> spex.soft_deleted_tables
    {'test_users'}
    >>> str(ListTablesAction(spex,ListTables()))
    "SQLExecutionResult({'success': True, 'data': []})"

    Create and delete with commited table that's soft deleted:
    >>> _ = CreateTableAction(spex, ct)
    >>> spex.temp_tables
    {'test_users'}
    >>> spex.soft_deleted_tables
    {'test_users'}

    >>> ist = easy_db.Table.from_existing('INFORMATION_SCHEMA.TABLES').run(spex)

    Confirm permanent table still exists
    >>> len(ist.select('table_name').where(ist.TABLE_NAME == 'test_users').run(spex))
    1

    #### Confirm temporary table exists ####
    >>> tmp = easy_db.Table.from_existing(make_temp_name(ct.table_name)).run(spex)
    >>> tmp.count().run(spex)
    0

    >>> _ = spex.rollback()

    Confirm permanent table still exists
    >>> len(ist.select('table_name').where(ist.TABLE_NAME == 'test_users').run(spex))
    1
    >>> spex.temp_tables
    set()
    >>> spex.soft_deleted_tables
    set()

    Confirm temporary table gone
    >>> try:
    ...     tmp = easy_db.Table.from_existing(make_temp_name(ct.table_name)).run(bex)
    ... except Exception as e:
    ...     print(e)
    {'error_code': 1146, 'error_message': "Table 'seneca_test.$temp$test_users' doesn't exist"}


    #### TEST: Write to temporary table ####
    Setup/clear state:
    >>> _ = spex.rollback(); spex.soft_deleted_tables |  spex.temp_tables
    set()
    >>> _ = spex.cur.execute('DROP DATABASE IF EXISTS seneca_test;')
    >>> _ = spex.cur.execute('CREATE DATABASE seneca_test;')
    >>> _ = spex.cur.execute('use seneca_test;')

    Create temporary table:
    >>> _ = CreateTableAction(spex, ct)
    >>> spex.temp_tables
    {'test_users'}
    >>> tmp = easy_db.Table.from_existing(ct.table_name).run(spex)
    >>> tmp.count().run(spex)
    0
    >>> tmp.insert([{'username':'test_perm'}]).run(spex)
    {'last_row_id': 1, 'row_count': 1}
    >>> tmp.count().run(spex)
    1

    >>> tmp_with_tmp_name = easy_db.Table.from_existing(make_temp_name(ct.table_name)).run(bex)
    >>> tmp_with_tmp_name.count().run(bex)
    1
    >>> _ = spex.commit()

    >>> tmp.count().run(bex)
    1
    >>> tmp.count().run(spex)
    1

    '''
    import doctest, sys
    import seneca.engine.storage.mysql_intermediate as mysqli
    import seneca.engine.storage.easy_db as easy_db
    from typing import Tuple

    from seneca.engine.storage.mysql_executer import Executer as Raw_Executer
    spex,bex = deps_provider(Tuple[Executer, Raw_Executer])
    ex = spex


    ct = CreateTable(
          'test_users',
          AutoIncrementColumn('id'),
          [ ColumnDefinition('username', SQLType('VARCHAR', 30), True),
            ColumnDefinition('drivers_licence_unmber', SQLType('VARCHAR', 30), True),
            ColumnDefinition('first_name', SQLType('VARCHAR', 30), False),
            ColumnDefinition('balance', SQLType('BIGINT'), False),
    ])

    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



"""
Note: Not threadsafe

This Redisnap backend stores data in Python objects. It can be used as a
standalone backend, but it's primarily designed to be used inside the
transactional backend.

TODO: Convert camelcase to snake.
TODO: Add type annotations to everything.
TODO: If there are enough type check of existing values, just change to
decorator.

TODO: Custom exception types, important!

"""
from seneca.engine.storage.redisnap.commands import *
import seneca.engine.storage.redisnap.resp_types as rtype
from seneca.engine.storage.redisnap.backend_abc import ExecuterBase as abc_executer


class Executer(abc_executer):
    """
    >>> e = ExecuterBase(); print(e)
    ExecuterBase({})
    """

    def __init__(self):
        self.data = {}

    def __repr__(self):
        return 'ExecuterBase(%s)' % str(self.data)

    def purge(self):
        self.data = {}

    def exists(self, cmd):
        return cmd.key in self.data

    def type(self, cmd):
        t = type(self.get(Get(cmd.key)))
        if t in [RScalarInt, RScalarFloat]:
            return RScalar

        return t

    def asserttype(self, cmd):
        return isinstance(self.get(cmd), cmd.r_type)

    def get(self, cmd):
        try:
            ret = self.data[cmd.key]
            assert isinstance(ret, RScalar), 'FSR we got the wrong type!'
            return ret
        except KeyError:
            return RDoesNotExist()

    def setnr(self, cmd):
        self.data[cmd.key] = make_rscalar(cmd.value)


    def del_(self, cmd):
        self.data.pop(cmd.key)

    def incrbynr(self, cmd):
        old = self(Get(cmd.key))
        old_type = type(old)

        if issubclass(old_type, RDoesNotExist):
            self(SetNR(cmd.key, cmd.amount))
        elif issubclass(old_type, RScalarInt):
            old.value += cmd.amount
        else:
            raise RedisVauleTypeError('Existing value has wrong type.')

    def appendnr(self, cmd):
        old = self(Get(cmd.key))
        old_type = type(old)
        assert isinstance(old, RScalar)

        if issubclass(old_type, RDoesNotExist):
            self(SetNR(cmd.key, cmd.value))
        elif issubclass(old_type, RScalarInt) or issubclass(old_type, RScalarFloat):
            self(SetNR(cmd.key, str(old.value) + cmd.value))
        elif issubclass(old_type, RScalar):
            old.value += cmd.value
        else:
            raise Exception('Existing value has wrong type.')

    def hget(self, cmd):
        try:
            maybe_rhash = self.data[cmd.key]
            if isinstance(maybe_rhash, RHash):
                return maybe_rhash.data[cmd.field]
            else:
                # TODO: custom exception type
                raise RedisKeyTypeError('Existing value has wrong type.')
        except KeyError:
            return RDoesNotExist()

    def hsetnr(self, cmd):
        if cmd.key in self.data:
            old_val = self.data[cmd.key]
            if not isinstance(old_val, RHash):
                raise RedisKeyTypeError('Existing value has wrong type.')
            else:
                inner_dict = self.data[cmd.key].data
                inner_dict[cmd.field] = make_rscalar(cmd.value)
        else:
            self.data[cmd.key] = RHash({cmd.field: make_rscalar(cmd.value)})

    def hexists(self, cmd):

        try:
            maybe_rhash = self.data[cmd.key]
            if isinstance(maybe_rhash, RHash):
                return cmd.field in maybe_rhash.data
            else:
                raise RedisKeyTypeError('Existing value has wrong type.')
        except KeyError:
            return False

    def lindex(self, cmd):
        try:
            maybe_rlist = self.data[cmd.key]
            if isinstance(maybe_rlist, RList):
                if len(maybe_rlist.data) <= cmd.index:
                     return RDoesNotExist()
                else:
                    return maybe_rlist.data[cmd.index]
            else:
                raise RedisKeyTypeError('Existing value has wrong type.')
        except KeyError:
            return RDoesNotExist()

    def lset(self, cmd):
        try:
            maybe_rlist = self.data[cmd.key]
            if isinstance(maybe_rlist, RList):
                if len(maybe_rlist.data) <= cmd.index:
                    raise RedisListOutOfRange('Index out of range.')
                maybe_rlist.data[cmd.index] = cmd.value
            else:
                raise RedisKeyTypeError('Existing value has wrong type.')
        except KeyError:
            raise RedisKeyTypeError('Cannot LSet an nonexistent key.')

    def _push_nr_base(self, method_name, key, values):
        if key in self.data:
            old_val = self.data[key]
            if not isinstance(old_val, RList):
                raise RedisKeyTypeError('Existing value has wrong type.')
            else:
                getattr(old_val.data, method_name)(*map(make_rscalar, values))
        else:
            self.data[key] = RList(list(map(make_rscalar, values)))

    def lpushnr(self, cmd):
        return self._push_nr_base('appendleft', cmd.key, cmd.value[::-1])

    def rpushnr(self, cmd):
        return self._push_nr_base('append', cmd.key, cmd.value)

    def _pop_base(self, method_name, cmd):
        try:
            maybe_rlist = self.data[cmd.key]
            if isinstance(maybe_rlist, RList):
                if not maybe_rlist.data:
                    raise Exception("Inconsistent state: should never have an empty RList, panic!")
                if len(maybe_rlist.data) == 1:
                    # Redis deletes lists on last pop
                    self.data.pop(cmd.key)
                    return maybe_rlist.data[0]
                else:
                    return getattr(maybe_rlist.data, method_name)()
            else:
                # TODO: custom exception type
                raise RedisKeyTypeError('Existing value has wrong type.')
        except KeyError:
            return RDoesNotExist()


    def lpop(self, cmd):
        return self._pop_base('popleft', cmd)


    def rpop(self, cmd):
        return self._pop_base('pop', cmd)


    def zaddnr(self, cmd):
        if cmd.key in self.data:
            existing_sset = self.data[cmd.key]
            if not isinstance(existing_sset, RSortedSet):
                raise RedisKeyTypeError('Existing value has wrong type.')
            else:
                for m, s in cmd.members_and_scores.items():
                    existing_sset.add(s, m)
        else:
            self.data[cmd.key] = RSortedSet(list(map(swap, cmd.members_and_scores.items())))

    def zremnr(self, cmd):
        if cmd.key in self.data:
            existing_sset = self.data[cmd.key]
            if not isinstance(existing_sset, RSortedSet):
                raise RedisKeyTypeError('Existing value has wrong type.')
            else:
                for m in cmd.members:
                    existing_sset.rem(m)
                if not existing_sset:
                    self.data.pop(cmd.key)
        else:
            pass

    def zrevrangebyscore(self, cmd):
        if cmd.key in self.data:
            existing_sset = self.data[cmd.key]
            if not isinstance(existing_sset, RSortedSet):
                raise RedisKeyTypeError('Existing value has wrong type.')
            else:
                return existing_sset.rev_range_by_score(cmd.max, cmd.min, cmd.inclusive, cmd.with_scores)
        else:
            return []


    def zscore(self, cmd):
        if cmd.key in self.data:
            existing_sset = self.data[cmd.key]
            if not isinstance(existing_sset, RSortedSet):
                raise RedisKeyTypeError('Existing value has wrong type.')
            else:
                try:
                    return existing_sset.score(cmd.member)
                except KeyError as e:
                    return None
        else:
            return None


    def zincrbynr(self, cmd):
        if cmd.key in self.data:
            existing_sset = self.data[cmd.key]
            if not isinstance(existing_sset, RSortedSet):
                raise RedisKeyTypeError('Existing value has wrong type.')
            else:
                if cmd.member in existing_sset:
                    existing_sset.incr_by(cmd.amount, cmd.member)
                else:
                    existing_sset.add(cmd.amount, cmd.member)
        else:
            self.data[cmd.key] = RSortedSet([(cmd.amount, cmd.member)])


    def __call__(self, cmd):
        # TODO: Make sure this is efficient and generally okay.
        if isinstance(cmd, Del):
            return self.del_(cmd)
        return getattr(self, cmd.__class__.__name__.lower())(cmd)

    def contains_command_addr(self, cmd):
        """
        >>> ex.purge()
        >>> ex.contains_command_addr(Get('foo'))
        False
        >>> ex.contains_command_addr(HGet('foo', 'bar'))
        False

        >>> ex(SetNR('foo', 'bar'))
        >>> ex.contains_command_addr(Get('foo'))
        True

        >>> return_exception_tuple(ex.contains_command_addr, HGet('foo', 'bar'))
        ('RedisKeyTypeError', 'Existing value has wrong type.')

        >>> ex(Del('foo'))
        >>> ex(HSetNR('foo', 'bar', 'baz'))

        >>> ex.contains_command_addr(HGet('foo', 'bar'))
        True
        """
        if hasattr(cmd, 'field'):
            return self.hexists(cmd)
        else:
            return self.exists(cmd)


def run_tests(deps_provider):
    ex = Executer()
    from seneca.engine.util import return_exception_tuple

    import doctest, sys
    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



from seneca.engine.util import auto_set_fields, fst, snd, swap
from abc import ABCMeta, abstractmethod
import inspect
from collections import deque
from sortedcontainers import SortedSet


class RedisKeyTypeError(Exception):
    pass

class RedisVauleTypeError(Exception):
    pass

class RedisListOutOfRange(Exception):
    pass

class ReprIsConstructor(metaclass=ABCMeta):
    def __repr__(self):
        #return '<RESP ADDRESS (%s) %s>' % (self.__class__.__name__, str(self.__dict__))
        args = list(inspect.signature(self.__class__.__init__).parameters.keys())[1:]
        auto_constructor_args = ', '.join(map(repr, [getattr(self, a) for a in args]))
        return '%s(%s)' % (self.__class__.__name__, auto_constructor_args)


## Data types ##
class RESPType(ReprIsConstructor):
    @auto_set_fields
    def __init__(self):
        pass

# Scalar Types
# TODO: to_bytes() methods for RScalar data types.
class RScalar(RESPType):
    @auto_set_fields
    def __init__(self, value):
        pass

    def to_bytes(self):
        return str.encode(self.value)

class RScalarInt(RScalar):
    def to_bytes(self):
        return str.encode(str(self.value))

class RScalarFloat(RScalar):
    def to_bytes(self):
        return str.encode(str(self.value))

class RList(RESPType):
    def __init__(self, data_list):
        self.data = deque(data_list)

    def __repr__(self):
        return 'RList(%s)' % self.data

class RSet(RESPType): pass

class RSortedSet(RESPType):
    def __init__(self, score_member_pair_list):
        self._sorted_set = SortedSet(score_member_pair_list, key=fst)
        self._dict = dict(map(swap, score_member_pair_list))

    def __repr__(self):
        """
        >>> RSortedSet([(1, 'foo'), (2, 'bar')])
        RSortedSet([(1, 'foo'), (2, 'bar')])
        """
        return 'RSortedSet(%s)'  % str(list(sorted(map(swap, self._dict.items()), key=fst)))

    def __contains__(self, maybe_member):
        """
        >>> 'foo' in RSortedSet([(1, 'foo'), (2, 'bar')])
        True

        >>> 'baz' in RSortedSet([(1, 'foo'), (2, 'bar')])
        False
        """
        return maybe_member in self._dict

    def __len__(self):
        """
        >>> len(RSortedSet([(1, 'foo'), (2, 'bar')]))
        2

        >>> len(RSortedSet([]))
        0
        """
        return len(self._dict)

    def __bool__(self):
        """
        >>> bool(RSortedSet([(1, 'foo'), (2, 'bar')]))
        True

        >>> bool(RSortedSet([]))
        False
        """
        return bool(self._dict)

    def add(self, score, member):
        """
        >>> s = RSortedSet([])
        >>> s.add(1,'foo'); s
        RSortedSet([(1, 'foo')])

        >>> s.add(2,'bar'); s
        RSortedSet([(1, 'foo'), (2, 'bar')])
        """
        self._sorted_set.add((score, member))
        self._dict[member] = score

    def rem(self, member):
        """
        >>> s = RSortedSet([(1, 'foo'), (2, 'bar')])
        >>> s.rem('foo'); s
        RSortedSet([(2, 'bar')])
        >>> s.rem('bar'); s
        RSortedSet([])

        >>> s.rem('bar')

        """
        try:
            score = self._dict.pop(member)
            self._sorted_set.discard((score, member))
        except KeyError as e:
            pass

    def rev_range_by_score(self, max_:int, min_:int, inclusive=(True,True), with_scores=False):
        """
        TODO: add withscores
        ZRANGEBYSCORE zset (1 5
        Will return all members with 1 < score <= 5 while:

        >>> s = RSortedSet([(1, 'one'), (2, 'two'), (3, 'three')])
        >>> list(s.rev_range_by_score(None, None))
        ['three', 'two', 'one']

        >>> list(s.rev_range_by_score(2, 1))
        ['two', 'one']

        >>> list(s.rev_range_by_score(2, 1, (False, True)))
        ['one']

        >>> list(s.rev_range_by_score(2, 1, (False, False)))
        []
        """
        res = self._sorted_set.irange((min_, None), (max_, None), inclusive=swap(inclusive), reverse=True)
        if with_scores:
            return res
        else:
            return map(snd, res)

    def score(self, member):
        """
        >>> s = RSortedSet([(1, 'foo'), (2, 'bar')])
        >>> s.score('foo')
        1
        >>> s.score('bar')
        2

        >>> s.score('baz')
        """
        try:
            return self._dict[member]
        except KeyError:
            return None

    def incr_by(self, amount, member):
        """
        >>> s = RSortedSet([(1, 'foo'), (2, 'bar')])
        >>> s.incr_by(1, 'foo'); s.score('foo')
        2
        >>> s.incr_by(2, 'foo'); s.score('foo')
        4
        >>> s
        RSortedSet([(2, 'bar'), (4, 'foo')])

        >>> s.incr_by(2, 'new_member'); s.score('new_member')
        2
        """
        score = self._dict.pop(member, 0)
        self._sorted_set.discard((score, member))

        new_score = score + amount
        self._dict[member] = new_score
        self._sorted_set.add((new_score, member))


# Collections
class RHash(RESPType):
    @auto_set_fields
    def __init__(self, data):
        pass

class RDoesNotExist(RScalarInt, RScalarFloat, RHash, RList, RSet, RSortedSet):
    """
    In Redis nonexistent keys are fully polymorphic.
    These must be stored in addresses after we do a del, so reads don't fall through and create a spurious dependency.

    >>> RDoesNotExist()
    RDoesNotExist()

    >>> RScalar('a')
    RScalar('a')

    >>> RScalarInt(7)
    RScalarInt(7)

    """
    @auto_set_fields
    def __init__(self):
        pass

    def __repr__(_):
        return 'RDoesNotExist()'

    def to_bytes(self):
        pass

class DataDependancy(ReprIsConstructor):
    @auto_set_fields
    def __init__(self, key):
        pass

class ExistentialDependancy(ReprIsConstructor):
    pass

class GeneralTypeDependancy(DataDependancy):
    pass

class ExactTypeDependancy(DataDependancy):
    pass

class KeyReadDependancy(DataDependancy):
    pass

class HashFieldReadDependancy(DataDependancy):
    @auto_set_fields
    def __init__(self, key, field):
        pass

def make_rscalar(val):
    if val is None:
        return RDoesNotExist()
    if isinstance(val, int):
        return RScalarInt(val)
    elif isinstance(val, float):
        return RScalarFloat(val)
    else:
        try:
            i = int(val)
            return RScalarInt(i)
        except ValueError:
            pass

        try:
            f = float(val)
            return RScalarFloat(f)
        except ValueError:
            pass

        assert isinstance(val, str)
        return RScalar(val)

def from_resp_str(type_byte_str):
    d = { 'string': RScalar,
          'hash': RHash,
          'none': RDoesNotExist
          # TODO: Finish this.
        }
    assert type_byte_str in d
    return d[type_byte_str]

def run_tests(deps_provider):
    from seneca.engine.util import return_exception_tuple

    import doctest, sys
    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



'''
Utility functions and components to support creation of MySQL queries

def surround(s, x):
    return s + x + s


def in_quotes(x):
    return surround('\'', x)


def in_backticks(x):
    return surround('`', x)

def sql_escapes(s):
    # TODO: make sure everything I need is here.
    import re
    return re.sub("'", "''", s)

'''
from seneca.engine.util import auto_set_fields, fst, snd, swap
from enum import Enum
from seneca.libs.types import datetime
from dateutil.parser import parse as parse_time
from terminaltables import AsciiTable
import re
from functools import wraps
from pydoc import locate

from MySQLdb import escape_string as mysql_escape


# Python type factory representing fixed length strings, outputted classes have a set length
class FixedStr(object) :
    def __init__(self):
        raise NotImplementedError("FixedStr shouldn't be called directly, use Len function")

    @classmethod
    def Len(cls, l):
        class NewClass(cls):
            _max_len = l
            name = 'FixedStrLen%d' % l

            def __init__(self, str):
                assert len(str) <= self._max_len
                self.str = str

            def __str__(self):
                return self.str

        NewClass.__name__ = NewClass.name
        NewClass.__str__ = lambda:NewClass.name

        return NewClass


sql_python_type_alist = [ ('BIGINT', int),
                          ('VARCHAR', FixedStr),
                          ('TEXT', str),
                          ('DATETIME', datetime),
                          ('BOOLEAN', bool),
                          ('DOUBLE', float),
                          ('TEXT', list),
                          ('TEXT', set)
]

valid_mysql_types = [x[0] for x in sql_python_type_alist]
supported_python_types = [x[1].__name__ for x in sql_python_type_alist]
mysql_py_dict = dict(sql_python_type_alist)
py_mysql_dict = dict(map(swap, sql_python_type_alist))


class SQLType(object):
    '''This class will create objects representing sql types.'''
    def __init__(self, type_str, *args):
        assert type_str in valid_mysql_types
        if type_str == 'VARCHAR':
            assert len(args) == 1, 'Type VARCHAR requires a length.'
            self.sql_type = ('VARCHAR', args[0])
        else:
            assert not args
            self.sql_type = type_str

    @classmethod
    def from_python_type(cls, python_type):
        if issubclass(python_type, FixedStr):
            return cls('VARCHAR', python_type._max_len)
        else:
            return cls(py_mysql_dict[python_type])

    @classmethod
    def from_db_describe_str(cls, st_str):

        st_str_upper = st_str.upper()
        if st_str_upper in valid_mysql_types:
            return cls(st_str_upper)

        varchar_len = re.search(r'^VARCHAR\((.+)\)$', st_str_upper)
        if varchar_len:
            l = int(varchar_len.group(1))
            return cls('VARCHAR', l)

        if st_str_upper.lower() in ['bigint(20)',
                                    'bigint(20) unsigned',
                                    'bigint(21)',
                                    'bigint(21) unsigned',
            ]:
            return cls('BIGINT')


        raise Exception('Unknown sql type string: ' + st_str)




    def __str__(self):
        if type(self.sql_type) == tuple and self.sql_type[0] =='VARCHAR':
            return '%s(%d)' % (self.sql_type[0], self.sql_type[1])
        else:
            return self.sql_type


def none_handler_decorator(f):
    @wraps(f)
    def wrapper(val):
        if val is None:
            return 'NULL'
        else:
            return f(val)
    return wrapper


def get_py_to_sql_cast_func(py_type):

    casting_func_dict = {
      int: lambda x: str(x),
      float: lambda x: str(x),
      str: lambda x: "'%s'" % str(mysql_escape(x), 'utf-8'),
      datetime: lambda x:'\'%s\'' % x.strftime('%Y-%m-%d %H:%M:%S'),
      bool: lambda x: 'TRUE' if x else 'FALSE',
      set: lambda x: "'%s'" % ','.join(x),
      list: lambda x: "'%s'" % ','.join(list(x)),
      # set: lambda x: "'%s'" % ';'.join([
      #   '{},{}'.format(str(xi),type(xi).__name__) for xi in list(x)
      #   if type(xi).__name__ in supported_python_types
      #   ]),
      # list: lambda x: "'%s'" % ';'.join([
      #   '{},{}'.format(str(xi),type(xi).__name__) for xi in x
      #   if type(xi).__name__ in supported_python_types
      #   ])
    }

    casting_func_dict = {k: none_handler_decorator(f) for (k, f) in casting_func_dict.items()}

    if py_type in casting_func_dict.keys():
        return casting_func_dict[py_type]
    elif issubclass(py_type, FixedStr):
        return casting_func_dict[str] # XXX: same casting method to TEXT and to VARCHAR(x)
    elif py_type is type(None): # NOTE: This seems to be the correct way to access NoneType in Python3 :-(
        return lambda _: 'NULL'
    else:
        # TODO: custom exception types
        raise Exception('Unsupported type, cannot convert to SQL str:' + str(py_type))

def get_str_to_py_cast_func(str_type):
    '''
    This is bad, needs to be rewritten, and should not be casting keys to strings.
    '''
    casting_func_dict = {
      'int': lambda x: int(x),
      'float': lambda x: float(x),
      'str': lambda x: '\'%s\'' % str(mysql_escape(x), 'utf-8'),
      'datetime': lambda x: parse_time(x),
      'bool': lambda x: bool(x),
      'set': lambda x: set(x.split(',')),
      'list': lambda x: x.split(','),
      # 'list': lambda x: [
      #   parse_time(i.split(',')[0]) if
      #       i.split(',')[1] == 'datetime' else
      #       locate(i.split(',')[1])(i.split(',')[0])
      #       for i in x.split(';') if i.split(',')[1] in supported_python_types
      # ],
      # 'set': lambda x: [
      #   parse_time(i.split(',')[0]) if
      #       i.split(',')[1] == 'datetime' else
      #       locate(i.split(',')[1])(i.split(',')[0])
      #       for i in list(x).split(';') if i.split(',')[1] in supported_python_types
      # ],
    }
    return casting_func_dict[str_type]

def cast_py_to_sql(py_val):
    '''Convenience function'''
    return get_py_to_sql_cast_func(type(py_val))(py_val)


def escape_sql_pattern(raw_string):
    # TODO: Make sure this is exhaustive.
    return raw_string.replace('_', '\\_').replace('%', '\\%')

class TabularKVs():
    def __init__(self, keys, rows):
        self.keys = tuple(keys)
        self.key_type = type(keys[0])
        self.record_len = len(keys)
        self.rows = list([self.format_row(x) for x in rows])

    # TODO: def format_keys, should verify no duplicates and all keys are of same type
    # TODO: fast_init method that skips all unecessary steps

    @classmethod
    def from_dicts(cls, ds):
        k0 = list(ds[0].keys())
        tab_kv = cls(k0, [])
        for kv in ds:
            tab_kv.append(kv)
        return tab_kv

    def to_klist_vlists(self):
        return self.keys, self.rows

    def format_row(self, row):
        assert len(row) == self.record_len
        t = type(row)

        if t == tuple:
            return row
        elif t == list:
            return tuple(row)
        elif t == dict:
            return self.row_from_dict(row)
        else:
            raise Exception('Unsupported type.')

    def __len__(self):
        return len(self.rows)

    def __bool__(self):
        return bool(self.__len__())

    # TODO: make a nice efficient __iter__ with an interator
    def __iter__(self):
        return [self.row_to_dict(r) for r in self.rows].__iter__()

    def row_from_dict(self, d):
        assert sorted(d.keys()) == sorted(self.keys)
        return tuple([d[k] for k in self.keys])

    def append(self, v):
        self.rows.append(self.format_row(v))

    def row_to_dict(self, r):
        return dict(zip(self.keys, r))

    # TODO: Decide if the current __repr__ and __str__ are user friendly
    def __str__(self):
        #https://stackoverflow.com/questions/5909873/how-can-i-pretty-print-ascii-tables-with-python
        table_data = [list(self.keys)] + list(map(list, self.rows))
        table = AsciiTable(table_data)
        return table.table

    def __repr__(self):
        return '\n'.join(list([str(self.row_to_dict(x)) for x in self.rows]))

    def __setitem__(self, key, val):
        self.rows[key] = self.format_row(val)

    def __getitem__(self, key):
        return self.row_to_dict(self.rows[key])

    def select(self, *keys):
        indices = [self.keys.index(k) for k in keys]
        return [[r[i] for i in indices] for r in self.rows]


def run_tests(_):
    '''
    '''
    # TODO: Add tests
    import doctest, sys
    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



"""
This module is the core logic of Redisnap, transactions/snapshots/savepoints are
implemented here.

It provides a superset of the the other backend APIs, adding methods to control
creation of new transactions, commits, and rollbacks.
"""
import redis as rr
from seneca.engine.storage.redisnap.commands import *
import seneca.engine.storage.redisnap.local_backend as l_back
import seneca.engine.storage.redisnap.redis_backend as r_back
from collections import OrderedDict
from seneca.engine.util import auto_set_fields
from abc import ABCMeta, abstractmethod
from seneca.engine.storage.redisnap.backend_abc import ExecuterBase as abc_executer


class NoTransactionsInGroup(Exception):
    pass


class OpTracker(ReprIsConstructor):
    """
    >>> o = OpTracker()
    >>> print(o)
    OpTracker({})

    >>> o.add(Get('foo')); print(o)
    OpTracker({'foo': [Get('foo')]})
    """
    # TODO: implement for hash field addrs as well
    def __init__(self):
        self.data = {}

    def add(self, cmd):
        if hasattr(cmd, 'field'):
            raise NotImplementedError()

        if cmd.key not in self.data:
            self.data[cmd.key] = []

        self.data[cmd.key].append(cmd)

    # TODO: This function has to both implement fields and get a lot smarter.
    def contains_command_addr(self, cmd):
        if hasattr(cmd, 'field'):
            raise NotImplementedError()

        return cmd.key in self.data

    def __repr__(self):
        return 'OpTracker(%s)' % str(self.data)


class Transaction:
    """
    >>> t = Transaction(None, 'testing')
    >>> print(t)
    Transaction({'_local_executer': ExecuterBase({}), '_read_deps': OpTracker({}), '_typecheck_deps': OpTracker({}), '_redo_log': OpTracker({}), '_mutation_tracker': OpTracker({}), '_status': 'open', 'tag': 'testing', '_transaction_group': None})
    """
    def __init__(self, transaction_group, tag):
        #self._local_executer = TransactionalExecuter()
        self._local_executer = l_back.Executer()

        self._read_deps = OpTracker()
        self._typecheck_deps = OpTracker()
        self._redo_log = OpTracker()
        self._mutation_tracker = OpTracker()

        self._status = 'open'
        self.tag = tag
        self._transaction_group = transaction_group # Needed

    def __repr__(self):
        return 'Transaction(%s)' % str(self.__dict__)

    def set_status(self, status):
        assert status in ['open', 'closed', 'dirty'], 'Invalid status: ' + str(status)
        self._status = status

    def get_status(self):
        return self._status

    def _get_upstream(self):
        return self._transaction_group._get_upstream_executer(self)

    def appendwo(self, cmd):
        current = self.recursive_upstream_call(Get(cmd.key))

        if isinstance(current, RDoesNotExist):
            self._local_executer(cmd)
        else:
            self._local_executer(Set(cmd.key, current + cmd.value))

    def run_mutate_command(self, cmd):
        cmd_type = type(cmd)

        if cmd_type == AppendNR:
            return self.appendwo(cmd)
        else:
            raise NotImplementedError()

    def __call__(self, cmd):
        if isinstance(cmd, Write):
            '''
            Writes are the simplest case: add them to the redo_log, run locally
            '''
            self._redo_log.add(cmd)
            self._local_executer(cmd)


        elif isinstance(cmd, Mutate):
            '''
            If the data is already present locally, it can simply be edited.
            If not, we need to recurse to get the data to be mutated, then write
            locally, and also track the mutation so if there are any subsequent
            reads, they get registered as external reads.
            '''
            self._redo_log.add(cmd)

            if self._local_executer.contains_command_addr(cmd):
                self._local_executer(cmd)
            else:
                self._mutation_tracker.add(cmd)
                self.run_mutate_command(cmd)


        if isinstance(cmd, TypeCheck):
            '''
            Typechecks are also simple. The really difficult stuff is a result
            of lazily evaluated mutating operations, typechecks don't care about
            those because they don't change the type.
            '''
            if self._local_executer.contains_command_addr(cmd):
                return self._local_executer(cmd)
            else:
                self._typecheck_deps.add(cmd)
                return self.recursive_upstream_call(cmd)


        elif isinstance(cmd, Read):
            '''
            Reads recurse if data not present locally, lack of local data
            triggers creation of a read dependency. Also, data that's read and
            has been regiestered in the mutation tracker also triggers a read
            dependency.
            '''
            if not self._local_executer.contains_command_addr(cmd):
                self._read_deps.add(cmd)
                return self.recursive_upstream_call(cmd)
            else:
                if self._mutation_tracker.contains_command_addr(cmd):
                    self._read_deps.add(cmd)

                return self._local_executer(cmd)


    def recursive_upstream_call(self, cmd):
        # Note: There should never be and dependency/redo logging in the method.
        if self._local_executer.contains_command_addr(cmd):
            return self._local_executer(cmd)
        else:
            return self._get_upstream().recursive_upstream_call(cmd)



class RedisBackendWapper:
    def __init__(self, *args, **kwargs):
        self.wrapped = r_back.Executer(*args, **kwargs)

    def __call__(self, cmd):
        return self.wrapped(cmd)

    def recursive_upstream_call(self, cmd):
        return self(cmd)

    def __getattr__(self, attr):
        return getattr(self.wrapped, attr)


class TransactionGroup(abc_executer):
    def __init__(self, *args, **kwargs):
        self._redis_backend = RedisBackendWapper(*args, **kwargs)
        self._active_transaction = None
        self._transactions = []
        self._transactions_by_tag = {}

    def finalize_current_transaction(self):
        if self._get_transaction_index(self._active_transaction) == len(self._transactions) - 1:
            self._active_transaction.set_status('closed')
            self._active_transaction = None
        else:
            raise NotImplementedError()

    def _get_transaction_index(self, ex):
        return list(filter(lambda i_x: i_x[1] == ex, enumerate(self._transactions)))[0][0]

    def _get_upstream_executer(self, ex):
        i = self._get_transaction_index(ex)
        if i > 0:
            return self._transactions[i - 1]
        else:
            return self._redis_backend


    def start_new_transaction(self, tag, index=None):
        """
        Specify an index to start writing
        """
        index = index if index else len(self._transactions)

        assert index <= len(self._transactions)
        assert tag not in self._transactions_by_tag, 'Tags must be unique.'
        # TODO: assert no bad transactions before this one

        s = Transaction(self, tag)
        self._transactions.insert(index, s)

        self._active_transaction = s
        self._transactions_by_tag[s.tag] = s

    def start_new_transaction_before_tag(self, this_tag, other_tag):
        # find this_tag index
        # start_new_transaction(self, tag, index=None)
        raise NotImplementedError()

    def rework_transaction(self, tag):
        raise NotImplementedError()

    def get_pending_ops(self):
        ret = []
        for t in self._transactions:
            for cmd_list in t._redo_log.data.values():
                for c in cmd_list:
                    ret.append(c)

        return ret

    def commit_all_to_redis(self):
        for c in self.get_pending_ops():
            self._redis_backend(c)
        self._transactions = []


    def clear(self):
        raise NotImplementedError()

    def __call__(self, cmd):
        # Dissallow writes directly to Redis, reads and typechecks are okay.
        if not self._transactions:
            raise NoTransactionsInGroup('You must create a transaction before executing write commands.')

        return cmd.safe_run(self._active_transaction)


def run_tests(deps_provider):
    '''
    >>> ex1 = TransactionGroup(host='127.0.0.1', port=32768)
    >>> ex._redis_backend.purge()

    # Operations should fail before transaction has been created.
    >>> exception_type_name(ex, Get('foo'))
    'NoTransactionsInGroup'

    >>> ex.start_new_transaction('testing-trans-1')

    >>> ex(AppendNR('foo', 'bar'))
    >>> ex._active_transaction._read_deps.contains_command_addr(Get('foo'))
    False
    >>> ex._active_transaction._mutation_tracker.contains_command_addr(Get('foo'))
    True

    >>> ex(Get('foo'))
    RScalar('bar')
    >>> ex._active_transaction._read_deps.contains_command_addr(Get('foo'))
    True
    >>> ex._active_transaction._mutation_tracker.contains_command_addr(Get('foo'))
    True

    >>> ex(Del('foo'))
    >>> ex(Get('foo'))
    RDoesNotExist()


    ### Run the following test without a Redis backend ###
    >>> rbe = ex._redis_backend
    >>> ex._redis_backend = None


    # Test tries to fall through to Redis, but it has been removed
    >>> exception_type_name(ex, Get('baz'))
    'AttributeError'


    # Set key in transaction 'testing-trans-1'
    >>> ex(SetNR('foo', 'bar'))

    >>> ex._active_transaction._local_executer(Get('foo'))
    RScalar('bar')

    ### Add Redis backend to TransactionGroup for more tests ###
    >>> ex._redis_backend = rbe

    # Read falls through to Redis
    >>> ex(Get('baz'))
    RDoesNotExist()

    >>> ex.start_new_transaction('testing-trans-2')
    >>> ex._active_transaction.tag
    'testing-trans-2'

    >>> ex._get_upstream_executer(ex._active_transaction) != ex._active_transaction
    True

    >>> t1 = ex._get_upstream_executer(ex._active_transaction)
    >>> t1.tag
    'testing-trans-1'

    >>> type(ex._get_upstream_executer(t1)).__name__
    'RedisBackendWapper'

    >>> ex(Get('baz'))
    RDoesNotExist()

    # Same tag value, will fail.
    >>> exception_type_name(ex.start_new_transaction, 'testing-trans-1')
    'AssertionError'

    >>> ex.start_new_transaction('testing-trans-3')

    # Foo not in active
    >>> ex._active_transaction._local_executer(Get('foo'))
    RDoesNotExist()

    # Foo not in Redis
    >>> ex._redis_backend(Get('foo'))
    RDoesNotExist()

    # Foo comes from testing-trans-1
    >>> ex(Get('foo'))
    RScalar('bar')

    # Overwrite foo in testing-trans-2
    >>> ex(SetNR('foo', 'baz')); ex(Get('foo'))
    RScalar('baz')
    >>> ex._transactions[0](Get('foo'))
    RScalar('bar')

    >>> ex(AppendNR('foo', 'bar'))
    >>> ex(Get('foo'))
    RScalar('bazbar')

    >>> ex._active_transaction._mutation_tracker.contains_command_addr(Get('foo'))
    False
    >>> ex(Get('new_key_doesnt_exist'))
    RDoesNotExist()

    ### Test commit ordering ###
    >>> ex2 = TransactionGroup(host='127.0.0.1', port=32768)
    >>> ex2._redis_backend.purge()
    >>> ex2.start_new_transaction('a')
    >>> ex2(SetNR('foo', 'a'))
    >>> ex2(SetNR('foo', 'b'))
    >>> ex2(SetNR('foo', 'c'))
    >>> ex2(SetNR('foo', 'd'))
    >>> ex2.start_new_transaction('b')
    >>> ex2(SetNR('foo', 'e'))
    >>> ex2(SetNR('foo', 'f'))
    >>> ex2(SetNR('foo', 'g'))
    >>> ex2(SetNR('foo', 'h'))
    >>> _ = [print(x) for x in ex2.get_pending_ops()]
    SetNR('foo', 'a')
    SetNR('foo', 'b')
    SetNR('foo', 'c')
    SetNR('foo', 'd')
    SetNR('foo', 'e')
    SetNR('foo', 'f')
    SetNR('foo', 'g')
    SetNR('foo', 'h')
    >>> ex.commit_all_to_redis()
    '''

    import doctest, sys

    ex = TransactionGroup(host='127.0.0.1', port=32768)
    from seneca.engine.util import return_exception_tuple

    def exception_type_name(*args):
        try:
            return args[0](*args[1:])
        except Exception as e:
            return type(e).__name__


    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



import random, seneca
from os import getenv as env
from dotenv import load_dotenv


def load_env():
    load_dotenv(dotenv_path='{}/../docker/redis.env'.format(path), override=True)


def get_redis_port(port=None):

    if env('CIRCLECI'):
        return 6379

    return env('REDIS_PORT', 6379)


def get_redis_password(password=None):
    if env('CIRCLECI'):
        return ''

    return env('REDIS_PASSWORD', '')

path = seneca.__path__[0]
load_env()

MASTER_DB = 0
DB_OFFSET = 1
NUM_CACHES = 2
CODE_OBJ_MAX_CACHE = 64
